/*******************************************************************************
 meta.c

 ADSK 2017-11-15 New file
*******************************************************************************/ 

#include "lqt_private.h"
#include <string.h>
#include <stdlib.h>

#define LOG_DOMAIN "meta"

void quicktime_meta_init(quicktime_meta_t *meta)
{
  memset( meta, 0, sizeof( quicktime_meta_t ) );
}

static int read_hdlr( quicktime_t *file )
{
  char type[4];

  // Skip Versoin, Flags, Predefined
  quicktime_skip( file, 8 );

  // Accroding to the QTFF spec, we must verify the handler type is "mdta"
  quicktime_read_char32(file, type);
  return strncmp( type, "mdta", 4 );
}

static void read_keys(quicktime_t *file, quicktime_meta_t *meta,
                       const quicktime_atom_t *keys_atom)
{
  unsigned i;
  
  // Skip Version, Flags
  quicktime_skip( file, 4 );
  
  meta->num_entries = quicktime_read_int32( file );
  
  // Verify count
  const int64_t leaf_position = quicktime_position(file);
  for ( i = 0; i < meta->num_entries &&
               quicktime_position(file) < keys_atom->end; i++ )
  {
    const unsigned total_key_size = quicktime_read_int32( file );
    quicktime_skip( file, total_key_size - 4 );
  }
  if ( meta->num_entries != i ) {
    lqt_log(file, LQT_LOG_WARNING, LOG_DOMAIN,
            "read_keys: less keys (%u) than announced (%u)",
            i, meta->num_entries );
    meta->num_entries = i;
  }

  
  meta->entries = (quicktime_meta_item_t*) calloc( meta->num_entries,
                                                   sizeof(quicktime_meta_item_t) );
                                                   
  // Now perform actual reading
  quicktime_set_position( file, leaf_position );
  for ( i = 0; i < meta->num_entries &&
               quicktime_position(file) < keys_atom->end; i++ )
  {
    const long key_size = quicktime_read_int32( file ) - 8;
    /* Skip key namespace for now */
    quicktime_skip( file, 4 );
    meta->entries[i].key = (char*)malloc( key_size + 1 );
    quicktime_read_data( file, (uint8_t*)meta->entries[i].key, key_size );
    meta->entries[i].key[key_size] = '\0';
  }
}

static void read_ilst( quicktime_t *file, quicktime_meta_t *meta,
                       const quicktime_atom_t *ilst_atom)
{
  quicktime_atom_t item_atom;
  quicktime_atom_t data_atom;
  unsigned long a, b, c, d;
  unsigned item_idx;
  unsigned val_size;
  
  while(quicktime_position(file) < ilst_atom->end)
  {
    quicktime_atom_read_header(file, &item_atom);

    /* Item index is in the type... */
    a = item_atom.type[0];
    b = item_atom.type[1];
    c = item_atom.type[2];
    d = item_atom.type[3];
    item_idx = (a << 24) | (b << 16) | (c << 8) | d;
    
    if ( item_idx > meta->num_entries ) {
      lqt_log(file, LQT_LOG_WARNING, LOG_DOMAIN,
              "read_ilst: item key (%u) is out of keys range (%u). Skipped.",
              item_idx, meta->num_entries );
      continue;
    }
    
    quicktime_atom_read_header(file, &data_atom);
    if( quicktime_atom_is(&data_atom, "data")) {
      /* Skip Type indicator and Locale */
      quicktime_skip( file, 8 );
      val_size = data_atom.size - 16;
      meta->entries[ item_idx - 1 ].val_size = val_size;
      /* Add a nul terminator after blob for user convenience.
         But leave val_size set to original value. */
      meta->entries[ item_idx - 1 ].val = (char*)malloc( val_size + 1 );
      quicktime_read_data( file, (uint8_t*)meta->entries[ item_idx - 1 ].val, val_size );
      meta->entries[ item_idx - 1 ].val[ val_size ] = '\0';
    }
    
    /* According to QTFF spec, there can be multiple values for an item
     but we will only consider first for now. */
    quicktime_atom_skip( file, &item_atom );
  }
}

int quicktime_read_meta(quicktime_t *file, quicktime_meta_t *meta,
                        const quicktime_atom_t *meta_atom)
{
  quicktime_atom_t sub_atom;
  int res = -1;
  
  while(quicktime_position(file) < meta_atom->end)
  {
    quicktime_atom_read_header(file, &sub_atom);

    if(quicktime_atom_is(&sub_atom, "hdlr")) {
      if ( read_hdlr( file ) ) {
        // Unexpected hdlr atom. Skip meta atom altogether...
        break;
      }
      /* Presence of valid hdlr atom is our success criteria */
      res = 0;
    }
    else if(quicktime_atom_is(&sub_atom, "keys")) {
      read_keys(file, meta, &sub_atom);
    }
    else if(quicktime_atom_is(&sub_atom, "ilst")) {
      read_ilst(file, meta, &sub_atom);
    }
    quicktime_atom_skip(file, &sub_atom);
  }
  return res;
}

void quicktime_write_meta(quicktime_t *file, quicktime_meta_t *meta)
{
  int i;
  quicktime_atom_t atom, subatom, item_atom, data_atom;

  quicktime_atom_write_header(file, &atom, "meta");
  
  quicktime_atom_write_header(file, &subatom, "hdlr");
  quicktime_write_int32(file, 0); /* Version and flags */
  quicktime_write_int32(file, 0); /* Predefined */
  quicktime_write_char32(file, "mdta");
  quicktime_write_int32(file, 0); /* Reserved */
  quicktime_write_int32(file, 0); /* Reserved */
  quicktime_write_int32(file, 0); /* Reserved */
  quicktime_write_char(file, '\0'); /* Name */
  quicktime_atom_write_footer(file, &subatom);
  
  quicktime_atom_write_header(file, &subatom, "keys");
  quicktime_write_int32(file, 0); /* Version and flags */
  quicktime_write_int32(file, meta->num_entries); /* Count */
  for ( i = 0; i < meta->num_entries; i++ ) {
    unsigned keyLen = strlen( meta->entries[i].key );
    quicktime_write_int32(file, 8+keyLen); /* key size */
    quicktime_write_char32(file, "mdta"); /* key namespace */
    quicktime_write_data(file, meta->entries[i].key, keyLen);
  }
  quicktime_atom_write_footer(file, &subatom);
  
  quicktime_atom_write_header(file, &subatom, "ilst");
  for ( i = 1; i <= meta->num_entries; i++ )
  {
    /* Item index goes in the item atom type... */
    char keyIndex[4];
    keyIndex[0] = i >> 24 & 0xff;
    keyIndex[1] = i >> 16 & 0xff;
    keyIndex[2] = i >> 8 & 0xff;
    keyIndex[3] = i >> 0 & 0xff;
    quicktime_atom_write_header(file, &item_atom, keyIndex);
    
      quicktime_atom_write_header(file, &data_atom, "data");
      quicktime_write_char(file, 0); /* type indicator: well-known */
      quicktime_write_int24(file, 1); /* well-known type: utf-8 */
      quicktime_write_int32(file, 0); /* default locale */
      quicktime_write_data(file, meta->entries[i-1].val, meta->entries[i-1].val_size);
      quicktime_atom_write_footer(file, &data_atom);

    quicktime_atom_write_footer(file, &item_atom);
  }
  quicktime_atom_write_footer(file, &subatom);
   
  quicktime_atom_write_footer(file, &atom);
}

void quicktime_meta_delete(quicktime_meta_t *meta)
{
  int i;
  if ( meta ) {
    for ( i =0; i < meta->num_entries; i++ ) {
      free( meta->entries[i].key );
      free( meta->entries[i].val );
    }
    free( meta->entries );
    meta->entries = 0;
  }
}
