/*******************************************************************************
 util.c

 libquicktime - A library for reading and writing quicktime/avi/mp4 files.
 http://libquicktime.sourceforge.net

 Copyright (C) 2002 Heroine Virtual Ltd.
 Copyright (C) 2002-2007 Members of the libquicktime project.

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License as published by the Free
 Software Foundation; either version 2.1 of the License, or (at your option)
 any later version.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License along
 with this library; if not, write to the Free Software Foundation, Inc., 51
 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*******************************************************************************/ 

#include "lqt_private.h"
#include "lqt_fseek.h"
#include "workarounds.h"
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <math.h>

#ifndef HAVE_LRINT
#define lrint(x) ((long int)(x))
#endif

#define LOG_DOMAIN "util"

#ifdef LINUX
#define PRINT_INT64 "%ld"
#else
#define PRINT_INT64 "%lld"
#endif

#define LQT_MIN(a, b) ((a<b)?a:b)
#define LQT_ROUNDUP(a, b) ((a+b-1)/(b)*(b))
#define LQT_ROUNDDOWN(a, b) ((a)/(b)*(b))

static int64_t quicktime_get_file_length(const char *path)
{
      struct stat status;
      if(stat(path, &status))
              perror("quicktime_get_file_length stat:");
      return status.st_size;
}

int quicktime_file_open(quicktime_t *file, const char *path, int rd, int wr)
{
  int exists = 0;

  /* ADSK 2012-06-27 I/O callbacks */
  if ( file->io_callback && ( rd ^ wr ) ) {
    if ( !file->io_callback->open ) { return 1; }
    exists = (file->io_callback->open( file->io_data, path, wr, &file->total_length ) == 0);
    if(!exists && !wr)
    {
      lqt_log(file, LQT_LOG_ERROR, LOG_DOMAIN, "%s", file->io_callback->lastError(file->io_data));
      return 1;
    }
  }
  else {
	char flags[10];
	if(rd && (file->stream = fopen(path, "rb")))
	{
		exists = 1; 
		fclose(file->stream); 
	}

	if(rd && !wr) sprintf(flags, "rb");
	else
	if(!rd && wr) sprintf(flags, "wb");
	else
	if(rd && wr)
	{
		if(exists) 
			sprintf(flags, "rb+");
		else
			sprintf(flags, "wb+");
	}

	if(!(file->stream = fopen(path, flags)))
	{
		return 1;
	}
        if(rd && exists)
        {
	        file->total_length = quicktime_get_file_length(path);		
        }
  }

  if(wr) {
    // ADSK 2018-08-03 Optimize writing
    file->write_alignment =
        file->io_callback && file->io_callback->get_write_alignment
        ? file->io_callback->get_write_alignment( file->io_data )
        : getpagesize();
    
    // Ensure minimum for setting mdat atom size in presave_top_buffer
    if ( file->write_alignment < 64 ) { file->write_alignment = 64; } 
    
    return posix_memalign( &file->presave_buffer,
                           file->write_alignment,
                           QUICKTIME_PRESAVE );
  }
  file->presave_top_buffer = 0;
  
  return 0;
}

int quicktime_file_close(quicktime_t *file)
{
  /* ADSK 2012-06-27 I/O callbacks */
  if ( file->io_callback )
  {
    /* ADSK 2021-07-22 In an error situation where file cannot be written
                       succesfully, we can now clear the wr flag before 
                       closing and avoid more errors. */
    if(file->wr)
    {
      /* ADSK 2018-06-06 I/O callbacks - write */
      if ( file->io_callback->write ) {
        if ( file->presave_size )
        {
          // ADSK 2018-08-03 Clear direct I/O padding at end of file.
          const int padding = file->presave_position - file->total_length;
          if ( padding > 0 ) {
            memset( file->presave_buffer + file->presave_size - padding,
                    0, padding );
          }

          file->io_callback->write(
            file->io_data,
            file->presave_position - file->presave_size,
            file->presave_buffer,
            file->presave_size );

          file->presave_size = 0;
        }

        // ADSK 2018-08-03 Optimize writing behind in first page
        if ( file->presave_top_buffer )
        {
          file->io_callback->write(
            file->io_data,
            0, // offset = beginning of file
            file->presave_top_buffer,
            file->write_alignment );
        }
      }
    }
    
    if ( file->io_callback->close ) {
      file->io_callback->close( file->io_data );
    }
  }

  if(file->stream)
  {
    if(file->wr)
    {
      /* Flush presave buffer */
      if(file->presave_size)
      {
        quicktime_fseek(file, file->presave_position - file->presave_size);
        fwrite(file->presave_buffer, 1, file->presave_size, file->stream);
      }

      if ( file->presave_top_buffer )
      {
        quicktime_fseek(file, 0);
        fwrite(file->presave_top_buffer, 1, file->write_alignment, file->stream);
      }
    }

    fclose(file->stream);
    file->presave_size = 0;
    file->stream = 0;
  }
  
  return 0;
}

/* ADSK 2012-06-27 dead code
int64_t quicktime_ftell(quicktime_t *file)
{
	return file->ftell_position;
}
*/

int quicktime_fseek(quicktime_t *file, int64_t offset)
{
	file->ftell_position = offset;
	/* ADSK 2012-09-10 Do seek past end of file so read actually fails 
        if(offset > file->total_length || offset < 0) return 1;
        */
        if(offset < 0) return 1;

        /* ADSK 2012-11-30 Nothing to do if no open stream. Should never happen.*/
	if(file->stream && fseeko(file->stream, file->ftell_position, SEEK_SET))
	{
//		perror("quicktime_fseek fseeko");
		return 1;
	}
	return 0;
}

/* Read entire buffer from the preload buffer */
static int read_preload(quicktime_t *file, uint8_t *data, int64_t size)
{
	int64_t selection_start = 0;
	int64_t selection_end = 0;
	int64_t fragment_start = 0;
	int64_t fragment_len = 0;

	selection_start = file->file_position;
	selection_end = quicktime_add(file->file_position, size);

	fragment_start = file->preload_ptr + (selection_start - file->preload_start);
	while(fragment_start < 0) fragment_start += file->preload_size;
	while(fragment_start >= file->preload_size) fragment_start -= file->preload_size;

	while(selection_start < selection_end)
	{
		fragment_len = selection_end - selection_start;
		if(fragment_start + fragment_len > file->preload_size)
			fragment_len = file->preload_size - fragment_start;

		memcpy(data, file->preload_buffer + fragment_start, fragment_len);
		fragment_start += fragment_len;
		data += fragment_len;

		if(fragment_start >= file->preload_size) fragment_start = (int64_t)0;
		selection_start += fragment_len;
	}
	return 0;
}

/* ADSK 2012-06-27 Custom I/O support. New function.*/
int quicktime_get_data(quicktime_t *file, const uint8_t **data, unsigned * size)
{
  if ( file->preload_size ) {
    /* Region requested for loading */
    int64_t selection_start = file->file_position;
    int64_t selection_end = file->file_position + *size;

    if ( selection_start >= file->preload_start && 
         selection_start < file->preload_end &&
         selection_end <= file->preload_end &&
         selection_end > file->preload_start)
    {
      /* Entire range is in preload buffer */
      *data = file->preload_buffer + selection_start - file->preload_start;
    }
    else {
      /* Not supported for now. */
      lqt_log(file, LQT_LOG_ERROR, LOG_DOMAIN,
              "quicktime_get_data(" PRINT_INT64 "-" PRINT_INT64 ") outside preload region(" PRINT_INT64 "-" PRINT_INT64 ")!",
              selection_start, selection_end, file->preload_start, file->preload_end );
      return 0;
    }
  }
  else {
    if ( !file->io_callback || !file->io_callback->get ) {
      return 0;
    }

    if ( file->io_callback->get( file->io_data, file->file_position,
                                 data, size ) ) { return 0; }
  }
  file->file_position += *size;
  return 1;
}

int quicktime_read_data(quicktime_t *file, uint8_t *data, int64_t size)
  {
  int result = 0;
  
  /* ADSK 2012-06-27 This function returns the number of bytes read
                     in an int but size is an int64_t... */
  if ( size > 0x7FFFffff ) {
    printf( "libquicktime read overflow\n" );
    // return 0;
  }
  /* Return if we had an error before */
  else if(file->io_error || file->io_eof) {
    // return 0;
  }
  /* ADSK 2012-11-30 Nothing to do if size is zero. */
  else if ( size <= 0 ) {
    // return 0;
  }
  /* ADSK 2012-06-27 I/O callbacks */
  else if ( file->io_callback && !file->preload_size ) {
    if ( file->io_callback->read ) {
      result = file->io_callback->read( file->io_data, file->file_position,
                                        data, (unsigned)size );
    }
    // else return 0;
  }
  else if(!file->preload_size)
    {
    quicktime_fseek(file, file->file_position);
    result = fread(data, 1, size, file->stream);
    if(result < size)
      {
      file->io_error = ferror(file->stream);
      file->io_eof   = feof(file->stream);
      }
    file->ftell_position += size;
    }
  else
    {
    /* Region requested for loading */
    int64_t selection_start = file->file_position;
    int64_t selection_end = file->file_position + size;
    int64_t fragment_start, fragment_len;

    if(selection_end - selection_start > file->preload_size)
      {
      /* Size is larger than preload size.  Should never happen. */
      /* ADSK 2013-05-08 Return error when file->stream is NULL. */
      if ( file->stream ) {
        quicktime_fseek(file, file->file_position);
        result = fread(data, 1, size, file->stream);
        if(result < size)
          {
          file->io_error = ferror(file->stream);
          file->io_eof   = feof(file->stream);
          }
        file->ftell_position += size;
      }
      // else return 0;
      }
    else if(selection_start >= file->preload_start && 
            selection_start < file->preload_end &&
            selection_end <= file->preload_end &&
            selection_end > file->preload_start)
      {
      /* Entire range is in buffer */
      read_preload(file, data, size);
      result = size;
      }
    else if(selection_end > file->preload_end && 
            selection_end - file->preload_size < file->preload_end)
      {
      /* Range is after buffer */
      /* Move the preload start to within one preload length of the selection_end */
      while(selection_end - file->preload_start > file->preload_size)
        {
        fragment_len = selection_end - file->preload_start - file->preload_size;
        if(file->preload_ptr + fragment_len > file->preload_size) 
          fragment_len = file->preload_size - file->preload_ptr;
        file->preload_start += fragment_len;
        file->preload_ptr += fragment_len;
        if(file->preload_ptr >= file->preload_size) file->preload_ptr = 0;
        }
      
      /* Append sequential data after the preload end to the new end */
      fragment_start = file->preload_ptr + file->preload_end - file->preload_start;
      while(fragment_start >= file->preload_size) 
        fragment_start -= file->preload_size;
      
      while(file->preload_end < selection_end)
        {
        fragment_len = selection_end - file->preload_end;
        if(fragment_start + fragment_len > file->preload_size)
          fragment_len = file->preload_size - fragment_start;
        quicktime_fseek(file, file->preload_end);
        result = fread(&(file->preload_buffer[fragment_start]),
                       fragment_len, 1, file->stream);

        if(result < fragment_len)
          {
          file->io_error = ferror(file->stream);
          file->io_eof   = feof(file->stream);
          }
        file->ftell_position += fragment_len;
        file->preload_end += fragment_len;
        fragment_start += fragment_len;
        if(fragment_start >= file->preload_size)
          fragment_start = 0;
        }
      
      read_preload(file, data, size);
      result = size;
      }
    else
      {
      /* Range is before buffer or over a preload_size away from the end of the buffer. */
      /* Replace entire preload buffer with range. */
      quicktime_fseek(file, file->file_position);
      result = fread(file->preload_buffer, 1, size, file->stream);
      if(result < size)
        {
        file->io_error = ferror(file->stream);
        file->io_eof   = feof(file->stream);
        }
      file->ftell_position += size;
      file->preload_start = file->file_position;
      file->preload_end = file->file_position + size;
      file->preload_ptr = 0;
      read_preload(file, data, size);
      

      }
    }
  if ( result ) {
    file->file_position += size;
    }
  else if ( size > 0 ) {
    /* ADSK 2012-09-10 Don't return garbage which may induce very nasty results */
    lqt_log(file, LQT_LOG_WARNING, LOG_DOMAIN,
            "Failed reading " PRINT_INT64 " bytes at offset " PRINT_INT64 ".",
            size, file->file_position );
    memset( data, 0, size );
    }
  return result;
  }

static size_t flush_presave( quicktime_t *file, int doPad )
{
  size_t writes_succeeded = 0;

  /* ADSK 2018-06-06 I/O callbacks - write */
  const int64_t filePos = file->presave_position - file->presave_size;
  if ( file->io_callback && file->io_callback->write )
  {
    if ( doPad ) {
      // ADSK 2018-08-03 Clear direct I/O padding at end of file.
      const int padding = file->presave_position - file->total_length;
      if ( padding > 0 ) {
        memset( file->presave_buffer + file->presave_size - padding,
                0, padding );
      }
    }

    writes_succeeded = file->io_callback->write(
        file->io_data,
        filePos,
        file->presave_buffer,
        (unsigned)file->presave_size );
  } else {
    quicktime_fseek(file, filePos);
    writes_succeeded = fwrite(file->presave_buffer, 1, file->presave_size, file->stream);
  }

  // ADSK 2018-08-03 Keep copy of first page
  if ( filePos < file->write_alignment ) {
    int numBytes = file->write_alignment - filePos;
    if ( numBytes > file->presave_size ) { numBytes = file->presave_size; }
    if ( ! file->presave_top_buffer ) {
      posix_memalign( &file->presave_top_buffer, file->write_alignment, file->write_alignment );
    }
    memcpy( file->presave_top_buffer + filePos, file->presave_buffer, numBytes );
  }

  file->presave_size = 0;
  
  return writes_succeeded;
}

int quicktime_write_data(quicktime_t *file, const uint8_t *data, int size)
  {
  const int reqSize = size;
  int data_offset = 0;
  int writes_attempted = 0;
  int writes_succeeded = 0;

  if(file->io_error)
    return 0;
  
  if(file->file_position != file->presave_position)
  {
    // ADSK 2018-08-03 Make better use of presave buffer when writing behind
    // (e.g. when writing atom size). No need to flush to disk!
    //
    if( file->presave_size && file->file_position < file->presave_position &&
        file->presave_position - file->file_position <= file->presave_size )
    {
      int overflow;
      const int offsetInBuf = file->presave_size + file->file_position
                              - file->presave_position;
      int fragment_size = QUICKTIME_PRESAVE - offsetInBuf;
      if ( fragment_size > size ) { fragment_size = size; }
      overflow = offsetInBuf + fragment_size - file->presave_size;

      memcpy(file->presave_buffer + offsetInBuf, data, fragment_size);

      size -= fragment_size;
      data_offset += fragment_size;
      if ( overflow > 0 ) {
        // We wrote into presave_buffer past presave_size.
        // Update presave_size but maintain alignment if applicable
        if ( file->io_callback && file->io_callback->write ) { 
          overflow = LQT_ROUNDUP( overflow, file->write_alignment );
        }
        file->presave_size += overflow;
        file->presave_position += overflow;
      }
    }
    else if ( file->presave_top_buffer &&
              file->file_position + size < file->write_alignment )
    {
      // ADSK 2018-08-03 Optimize writing behind in first page
      memcpy(file->presave_top_buffer + file->file_position, data, size);
      /* Adjust file position */
      file->file_position += size;
      file->ftell_position += size;
      return 1;
    }
    else
    {
      // We must map another section of file with presave_buffer...
      
      if(file->presave_size)
      {
        // Flush existing buffer
        writes_succeeded += flush_presave( file, 1 /* doPad */ );
      }
      
      // ADSK 2018-08-03 Optimize writing behind in other pages. 
      if( file->io_callback && file->io_callback->read )
      {
        file->presave_position = LQT_ROUNDDOWN( file->file_position,
                                                file->write_alignment );
        do {
          file->presave_size += file->write_alignment;
          file->presave_position += file->write_alignment;
        } while ( file->file_position + size + file->write_alignment
                  > file->presave_position );
        
        // In case we are writing behind, let's read in previously written data
        // in our presave buffer. We will be overwriting small portions of it
        // and will later write those blocks back.
        const int64_t readOffset = file->presave_position - file->presave_size;
        const int readResult = file->io_callback->read(
                                  file->io_data, readOffset,
                                  file->presave_buffer, file->presave_size );
        if ( readResult < LQT_MIN( file->presave_size,
                                   file->total_length - readOffset ) )
        {
          lqt_log(file, LQT_LOG_ERROR, LOG_DOMAIN,
              "quicktime_write_data(" PRINT_INT64 "-" PRINT_INT64 ") fails "
              "reading prewritten data(" PRINT_INT64 "-" PRINT_INT64 "): %d",
              file->file_position, file->file_position+size,
              readOffset, file->presave_position, readResult );
          return 0;
        }
        
        return quicktime_write_data(file, data, size);
      }
      // Seek to new position
      else {
        file->presave_position = file->file_position;
      }
    }
  }
  
  // Write presave buffers until done
  while(size > 0)
  {
    int fragment_size = QUICKTIME_PRESAVE;
    if(fragment_size > size) fragment_size = size;
    if(fragment_size + file->presave_size > QUICKTIME_PRESAVE)
      fragment_size = QUICKTIME_PRESAVE- file->presave_size;

    memcpy(file->presave_buffer + file->presave_size,
           data + data_offset,
           fragment_size);

    if ( file->io_callback && file->io_callback->write && 
         file->presave_position >= file->total_length )
    {
      // ADSK 2018-08-03 Pad last block for direct I/O
      const int padded = LQT_ROUNDUP( fragment_size, file->write_alignment);
      file->presave_position += padded;
      file->presave_size += padded;
    }
    else {
      file->presave_position += fragment_size;
      file->presave_size += fragment_size;
    }
    data_offset += fragment_size;
    size -= fragment_size;

    if(size > 0 && file->presave_size >= QUICKTIME_PRESAVE)
    {
      writes_succeeded += flush_presave( file, 0 /* doPad */ );
    }
  }

  /* Adjust file position */
  file->file_position += reqSize;
  /* Adjust ftell position */
  file->ftell_position = file->file_position;
  /* Adjust total length */
  if(file->total_length < file->ftell_position) file->total_length = file->ftell_position;

  /* fwrite failed */
  if(!writes_succeeded && writes_attempted && file->stream)
    {
    file->io_error = ferror(file->stream);
    return 0;
    }
  else
    if(!size)
      return 1;
    else
      return size;
  }

int64_t quicktime_byte_position(quicktime_t *file)
{
	return quicktime_position(file);
}


void quicktime_read_pascal(quicktime_t *file, char *data)
{
	char len = quicktime_read_char(file);
	quicktime_read_data(file, (uint8_t*)data, len);
	data[(int)len] = 0;
}

void quicktime_write_pascal(quicktime_t *file, char *data)
{
	uint8_t len = strlen(data);
	quicktime_write_data(file, &len, 1);
	quicktime_write_data(file, (uint8_t*)data, len);
}

float quicktime_read_fixed32(quicktime_t *file)
{
	unsigned long a, b, c, d;
	uint8_t data[4];

	quicktime_read_data(file, data, 4);
	a = data[0];
	b = data[1];
	c = data[2];
	d = data[3];
	
	a = (a << 8) + b;
	b = (c << 8) + d;

	if(b)
		return (float)a + (float)b / 65536;
	else
		return a;
}


int quicktime_write_fixed32(quicktime_t *file, float number)
{
	unsigned char data[4];
	int a, b;

	a = number;
	b = (number - a) * 65536;
	data[0] = a >> 8;
	data[1] = a & 0xff;
	data[2] = b >> 8;
	
	data[3] = b & 0xff;

	return quicktime_write_data(file, data, 4);
}

static float
float32_be_read (unsigned char *cptr)
{       int             exponent, mantissa, negative ;
        float   fvalue ;

        negative = cptr [0] & 0x80 ;
        exponent = ((cptr [0] & 0x7F) << 1) | ((cptr [1] & 0x80) ? 1 : 0) ;
        mantissa = ((cptr [1] & 0x7F) << 16) | (cptr [2] << 8) | (cptr [3]) ;

        if (! (exponent || mantissa))
                return 0.0 ;

        mantissa |= 0x800000 ;
        exponent = exponent ? exponent - 127 : 0 ;

        fvalue = mantissa ? ((float) mantissa) / ((float) 0x800000) : 0.0 ;

        if (negative)
                fvalue *= -1 ;

        if (exponent > 0)
                fvalue *= (1 << exponent) ;
        else if (exponent < 0)
                fvalue /= (1 << abs (exponent)) ;

        return fvalue ;
} /* float32_be_read */

static double
double64_be_read (unsigned char *cptr)
{       int             exponent, negative ;
        double  dvalue ;

        negative = (cptr [0] & 0x80) ? 1 : 0 ;
        exponent = ((cptr [0] & 0x7F) << 4) | ((cptr [1] >> 4) & 0xF) ;

        /* Might not have a 64 bit long, so load the mantissa into a double. */
        dvalue = (((cptr [1] & 0xF) << 24) | (cptr [2] << 16) | (cptr [3] << 8) | cptr [4]) ;
        dvalue += ((cptr [5] << 16) | (cptr [6] << 8) | cptr [7]) / ((double) 0x1000000) ;

        if (exponent == 0 && dvalue == 0.0)
                return 0.0 ;

        dvalue += 0x10000000 ;

        exponent = exponent - 0x3FF ;

        dvalue = dvalue / ((double) 0x10000000) ;

        if (negative)
                dvalue *= -1 ;

        if (exponent > 0)
                dvalue *= (1 << exponent) ;
        else if (exponent < 0)
                dvalue /= (1 << abs (exponent)) ;

        return dvalue ;
} /* double64_be_read */

static void
float32_be_write (float in, unsigned char *out)
{       int             exponent, mantissa, negative = 0 ;

        memset (out, 0, sizeof (int)) ;

        if (in == 0.0)
                return ;

        if (in < 0.0)
        {       in *= -1.0 ;
                negative = 1 ;
                } ;

        in = frexp (in, &exponent) ;

        exponent += 126 ;

        in *= (float) 0x1000000 ;
        mantissa = (((int) in) & 0x7FFFFF) ;

        if (negative)
                out [0] |= 0x80 ;

        if (exponent & 0x01)
                out [1] |= 0x80 ;

        out [3] = mantissa & 0xFF ;
        out [2] = (mantissa >> 8) & 0xFF ;
        out [1] |= (mantissa >> 16) & 0x7F ;
        out [0] |= (exponent >> 1) & 0x7F ;

        return ;
} /* float32_be_write */


static void
double64_be_write (double in, unsigned char *out)
{       int             exponent, mantissa ;

        memset (out, 0, sizeof (double)) ;

        if (in == 0.0)
                return ;

        if (in < 0.0)
        {       in *= -1.0 ;
                out [0] |= 0x80 ;
                } ;

        in = frexp (in, &exponent) ;

        exponent += 1022 ;

        out [0] |= (exponent >> 4) & 0x7F ;
        out [1] |= (exponent << 4) & 0xF0 ;

        in *= 0x20000000 ;
        mantissa = lrint (floor (in)) ;

        out [1] |= (mantissa >> 24) & 0xF ;
        out [2] = (mantissa >> 16) & 0xFF ;
        out [3] = (mantissa >> 8) & 0xFF ;
        out [4] = mantissa & 0xFF ;

        in = fmod (in, 1.0) ;
        in *= 0x1000000 ;
        mantissa = lrint (floor (in)) ;

        out [5] = (mantissa >> 16) & 0xFF ;
        out [6] = (mantissa >> 8) & 0xFF ;
        out [7] = mantissa & 0xFF ;

        return ;
} /* double64_be_write */


float quicktime_read_float32(quicktime_t *file)
  {
  unsigned char b[4];
  quicktime_read_data(file, b, 4);
  return float32_be_read(b);
  }

double quicktime_read_double64(quicktime_t *file)
  {
  unsigned char b[8];
  quicktime_read_data(file, b, 8);
  return double64_be_read(b);
  }


int quicktime_write_float32(quicktime_t *file, float value)
  {
  unsigned char b[4];
  float32_be_write(value, b);
  return quicktime_write_data(file, b, 4);
  }

int quicktime_write_double64(quicktime_t *file, double value)
  {
  unsigned char b[8];
  double64_be_write(value, b);
  return quicktime_write_data(file, b, 8);
  }

int quicktime_write_int64(quicktime_t *file, int64_t value)
{
	unsigned char data[8];

	data[0] = (((uint64_t)value) & 0xff00000000000000LL) >> 56;
	data[1] = (((uint64_t)value) & 0xff000000000000LL) >> 48;
	data[2] = (((uint64_t)value) & 0xff0000000000LL) >> 40;
	data[3] = (((uint64_t)value) & 0xff00000000LL) >> 32;
	data[4] = (((uint64_t)value) & 0xff000000LL) >> 24;
	data[5] = (((uint64_t)value) & 0xff0000LL) >> 16;
	data[6] = (((uint64_t)value) & 0xff00LL) >> 8;
	data[7] =  ((uint64_t)value) & 0xff;

	return quicktime_write_data(file, data, 8);
}

int quicktime_write_int64_le(quicktime_t *file, int64_t value)
{
        unsigned char data[8];
 
        data[7] = (((uint64_t)value) & 0xff00000000000000LL) >> 56;
        data[6] = (((uint64_t)value) & 0xff000000000000LL) >> 48;
        data[5] = (((uint64_t)value) & 0xff0000000000LL) >> 40;
        data[4] = (((uint64_t)value) & 0xff00000000LL) >> 32;
        data[3] = (((uint64_t)value) & 0xff000000LL) >> 24;
        data[2] = (((uint64_t)value) & 0xff0000LL) >> 16;
        data[1] = (((uint64_t)value) & 0xff00LL) >> 8;
        data[0] =  ((uint64_t)value) & 0xff;
                                                                                                                  
        return quicktime_write_data(file, data, 8);
}


int quicktime_write_int32(quicktime_t *file, long value)
{
	unsigned char data[4];

	data[0] = (value & 0xff000000) >> 24;
	data[1] = (value & 0xff0000) >> 16;
	data[2] = (value & 0xff00) >> 8;
	data[3] = value & 0xff;

	return quicktime_write_data(file, data, 4);
}

int quicktime_write_int32_le(quicktime_t *file, long value)
{
	unsigned char data[4];

	data[3] = (value & 0xff000000) >> 24;
	data[2] = (value & 0xff0000) >> 16;
	data[1] = (value & 0xff00) >> 8;
	data[0] = value & 0xff;

	return quicktime_write_data(file, data, 4);
}

int quicktime_write_char32(quicktime_t *file, char *string)
{
return quicktime_write_data(file, (uint8_t*)string, 4);
}


float quicktime_read_fixed16(quicktime_t *file)
{
	unsigned char data[2];
	
	quicktime_read_data(file, data, 2);
	if(data[1])
		return (float)data[0] + (float)data[1] / 256;
	else
		return (float)data[0];
}

int quicktime_write_fixed16(quicktime_t *file, float number)
{
	unsigned char data[2];
	int a, b;

	a = number;
	b = (number - a) * 256;
	data[0] = a;
	data[1] = b;

	return quicktime_write_data(file, data, 2);
}

unsigned long quicktime_read_uint32(quicktime_t *file)
{
	unsigned long result;
	unsigned long a, b, c, d;
	uint8_t data[4];

	quicktime_read_data(file, data, 4);
	a = data[0];
	b = data[1];
	c = data[2];
	d = data[3];

	result = (a << 24) | (b << 16) | (c << 8) | d;
	return result;
}

long quicktime_read_int32(quicktime_t *file)
{
  /* ADSK 2013-08-27 : Allow returning negative long in 64bit */
	long result, a;
	unsigned long b, c, d;
	uint8_t data[4];

	quicktime_read_data(file, data, 4);
	a = (int8_t)data[0];
	b = data[1];
	c = data[2];
	d = data[3];

	result = (a << 24) | (b << 16) | (c << 8) | d;
	return result;
}

long quicktime_read_int32_le(quicktime_t *file)
{
	unsigned long result;
	unsigned long a, b, c, d;
	uint8_t data[4];

	quicktime_read_data(file, data, 4);
	a = data[0];
	b = data[1];
	c = data[2];
	d = data[3];

	result = (d << 24) | (c << 16) | (b << 8) | a;
	return (long)result;
}

int64_t quicktime_read_int64(quicktime_t *file)
{
	uint64_t result, a, b, c, d, e, f, g, h;
	uint8_t data[8];

	quicktime_read_data(file, data, 8);
	a = data[0];
	b = data[1];
	c = data[2];
	d = data[3];
	e = data[4];
	f = data[5];
	g = data[6];
	h = data[7];

	result = (a << 56) | 
		(b << 48) | 
		(c << 40) | 
		(d << 32) | 
		(e << 24) | 
		(f << 16) | 
		(g << 8) | 
		h;
	return (int64_t)result;
}


int64_t quicktime_read_int64_le(quicktime_t *file)
{
        uint64_t result, a, b, c, d, e, f, g, h;
        uint8_t data[8];
                                                                                                                  
        quicktime_read_data(file, data, 8);
        a = data[7];
        b = data[6];
        c = data[5];
        d = data[4];
        e = data[3];
        f = data[2];
        g = data[1];
        h = data[0];
                                                                                                                  
        result = (a << 56) |
                (b << 48) |
                (c << 40) |
                (d << 32) |
                (e << 24) |
                (f << 16) |
                (g << 8) |
                h;
        return (int64_t)result;
}


long quicktime_read_int24(quicktime_t *file)
{
	unsigned long result;
	unsigned long a, b, c;
        uint8_t data[4];
	
	quicktime_read_data(file, data, 3);
	a = data[0];
	b = data[1];
	c = data[2];

	result = (a << 16) | (b << 8) | c;
	return (long)result;
}

int quicktime_write_int24(quicktime_t *file, long number)
{
	unsigned char data[3];
	data[0] = (number & 0xff0000) >> 16;
	data[1] = (number & 0xff00) >> 8;
	data[2] = (number & 0xff);
	
	return quicktime_write_data(file, data, 3);
}

int quicktime_read_int16(quicktime_t *file)
{
	unsigned long result;
	unsigned long a, b;
        uint8_t data[2];
	
	quicktime_read_data(file, data, 2);
	a = data[0];
	b = data[1];

	result = (a << 8) | b;
	return (int)result;
}

int quicktime_read_int16_le(quicktime_t *file)
{
	unsigned long result;
	unsigned long a, b;
	uint8_t data[2];
	
	quicktime_read_data(file, data, 2);
	a = data[0];
	b = data[1];

	result = (b << 8) | a;
	return (int)result;
}

int quicktime_write_int16(quicktime_t *file, int number)
{
	unsigned char data[2];
	data[0] = (number & 0xff00) >> 8;
	data[1] = (number & 0xff);
	
	return quicktime_write_data(file, data, 2);
}

int quicktime_write_int16_le(quicktime_t *file, int number)
{
	unsigned char data[2];
	data[1] = (number & 0xff00) >> 8;
	data[0] = (number & 0xff);
	
	return quicktime_write_data(file, data, 2);
}

int quicktime_read_char(quicktime_t *file)
{
	char output;
	quicktime_read_data(file, (uint8_t*)(&output), 1);
	return output;
}

int quicktime_write_char(quicktime_t *file, char x)
{
return quicktime_write_data(file, (uint8_t*)(&x), 1);
}

void quicktime_read_char32(quicktime_t *file, char *string)
{
quicktime_read_data(file, (uint8_t*)string, 4);
}

int64_t quicktime_position(quicktime_t *file) 
{ 
	return file->file_position; 
}

int quicktime_set_position(quicktime_t *file, int64_t position) 
{
	/* ADSK 2010-12-01: Reset the EOF flag when reseting the position */
	if ( file->file_position != position )
	{
		file->file_position = position;
		file->io_eof = 0;
	}
	return 0;
}

/* ADSK 2017-11-15 Add skip function */
void quicktime_skip(quicktime_t *file, unsigned num_bytes)
{
	quicktime_set_position(file, file->file_position + num_bytes );
}

void quicktime_copy_char32(char *output, char *input)
{
	*output++ = *input++;
	*output++ = *input++;
	*output++ = *input++;
	*output = *input;
}


void quicktime_print_chars(char *desc, uint8_t *input, int len)
{
	int i;
	lqt_dump("%s", desc);
	for(i = 0; i < len; i++) lqt_dump("%02x ", input[i]);
	lqt_dump("\n");
}

unsigned long quicktime_current_time(void)
{
	time_t t;
	time (&t);
	return (t+(66*31536000)+1468800);
}

int quicktime_match_32(void *_input, void *_output)
{
        uint8_t * input = (uint8_t*)_input;
        uint8_t * output = (uint8_t*)_output;
	if(input[0] == output[0] &&
		input[1] == output[1] &&
		input[2] == output[2] &&
		input[3] == output[3])
		return 1;
	else 
		return 0;
}

int quicktime_match_24(char *input, char *output)
{
	if(input[0] == output[0] &&
		input[1] == output[1] &&
		input[2] == output[2])
		return 1;
	else 
		return 0;
}

static void do_hexdump(uint8_t * data, int len, int linebreak, FILE * f)
  {
  int i;
  int bytes_written = 0;
  int imax;

  while(bytes_written < len)
    {
    imax = (bytes_written + linebreak > len) ? len - bytes_written : linebreak;
    for(i = 0; i < imax; i++)
      fprintf(f, "%02x ", data[bytes_written + i]);
    for(i = imax; i < linebreak; i++)
      fprintf(f, "   ");
    for(i = 0; i < imax; i++)
      {
      if(!(data[bytes_written + i] & 0x80) && (data[bytes_written + i] >= 32))
        fprintf(f, "%c", data[bytes_written + i]);
      else
        fprintf(f, ".");
      }
    bytes_written += imax;
    fprintf(f, "\n");
    }
  }

void lqt_hexdump(uint8_t * data, int len, int linebreak)
  {
  do_hexdump(data, len, linebreak, stderr);  
  }

void lqt_hexdump_stdout(uint8_t * data, int len, int linebreak)
  {
  do_hexdump(data, len, linebreak, stdout);  
  }

void lqt_dump_time(uint64_t t)
  {
  time_t ti;
  struct tm tm;
  /*  2082844800 = seconds between 1/1/04 and 1/1/70 */
  ti = t - 2082844800;
  localtime_r(&ti, &tm);
  tm.tm_mon++;
  tm.tm_mday++;
  printf("%04d-%02d-%02d %02d:%02d:%02d (%"PRId64")",
         tm.tm_year+1900, tm.tm_mon, tm.tm_mday,
         tm.tm_hour, tm.tm_min, tm.tm_sec, t);
  }
